#ifndef _EOLIAN_TYPEDEF_EO_STUB_H_STUBS
#define _EOLIAN_TYPEDEF_EO_STUB_H_STUBS

typedef Eo Typedef;

/** Docs for typedef Evas.Coord.
 *
 * @since 1.66
 *
 * @ingroup Evas_Coord
 */
typedef int Evas_Coord;

/** Docs for typedef List_Objects.
 *
 * @since 1.66
 *
 * @ingroup List_Objects
 */
typedef Eina_Iterator *List_Objects;

/** Docs for typedef Evas.Coord2.
 *
 * @since 1.66
 *
 * @ingroup Evas_Coord2
 */
typedef Evas_Coord Evas_Coord2;

/** Docs for typedef Evas.Coord3.
 *
 * @since 1.66
 *
 * @ingroup Evas_Coord3
 */
typedef Evas_Coord2 Evas_Coord3;


#endif
