/* sub mask x color -> dst */

static void
init_sub_mask_color_span_funcs_c(void)
{
}

static void
init_sub_mask_color_pt_funcs_c(void)
{
}

/*-----*/

/* sub_rel mask x color -> dst */

static void
init_sub_rel_mask_color_span_funcs_c(void)
{
}

static void
init_sub_rel_mask_color_pt_funcs_c(void)
{
}
