/* sub pixel x color --> dst */

static void
init_sub_pixel_color_span_funcs_c(void)
{
}

static void
init_sub_pixel_color_pt_funcs_c(void)
{
}

/*-----*/

/* sub_rel pixel x color --> dst */

static void
init_sub_rel_pixel_color_span_funcs_c(void)
{
}

static void
init_sub_rel_pixel_color_pt_funcs_c(void)
{
}
