/* sub pixel x mask --> dst */

/* XXX: not used
static void
init_sub_pixel_mask_span_funcs_c(void)
{
}
*/

/* XXX: not used
static void
init_sub_pixel_mask_pt_funcs_c(void)
{
}
*/

/*-----*/

/* sub_rel pixel x mask --> dst */

static void
init_sub_rel_pixel_mask_span_funcs_c(void)
{
}

static void
init_sub_rel_pixel_mask_pt_funcs_c(void)
{
}
