#ifndef _ELM_ACCESS_EO_H_
#define _ELM_ACCESS_EO_H_

#ifndef _ELM_ACCESS_EO_CLASS_TYPE
#define _ELM_ACCESS_EO_CLASS_TYPE

typedef Eo Elm_Access;

#endif

#ifndef _ELM_ACCESS_EO_TYPES
#define _ELM_ACCESS_EO_TYPES


#endif
/** Elm abstract accessibility class
 *
 * @ingroup Elm_Access
 */
#define ELM_ACCESS_CLASS elm_access_class_get()

EWAPI const Efl_Class *elm_access_class_get(void) EINA_CONST;

#endif
