#ifndef _ELDBUS_CXX_SUITE_H
#define _ELDBUS_CXX_SUITE_H

#include <cassert>
#include <algorithm>

#include <check.h>
#include "../efl_check.h"
void eldbus_test_connection(TCase* tc);
void eldbus_test_client(TCase* tc);

#endif /* _ELDBUS_CXX_SUITE_H */
