/* sub color -> dst */

#ifdef BUILD_MMX
static void
init_sub_color_span_funcs_mmx(void)
{
}
#endif

#ifdef BUILD_MMX
static void
init_sub_color_pt_funcs_mmx(void)
{
}
#endif

/*-----*/

/* sub_rel color -> dst */

#ifdef BUILD_MMX
static void
init_sub_rel_color_span_funcs_mmx(void)
{
}
#endif

#ifdef BUILD_MMX
static void
init_sub_rel_color_pt_funcs_mmx(void)
{
}
#endif
