#ifndef _ECORE_CXX_SUITE_H
#define _ECORE_CXX_SUITE_H

#include <cassert>
#include <algorithm>

#include <check.h>
#include "../efl_check.h"
void ecore_cxx_test_safe_call(TCase* tc);

#endif /* _ECORE_CXX_SUITE_H */
