int         e_prefix_determine(char *argv0);
void        e_prefix_shutdown(void);
const char *e_prefix_get(void);
const char *e_prefix_bin_get(void);
const char *e_prefix_data_get(void);
const char *e_prefix_lib_get(void);
