#ifdef HAVE_CONFIG_H
# include "elementary_config.h"
#endif

#define EFL_PACK_LAYOUT_PROTECTED

#include <Elementary.h>
#include "elm_priv.h"

typedef struct {

} Efl_Ui_Spotlight_Manager_Data;


#include "efl_ui_spotlight_manager.eo.c"
