#ifndef _ELM_SLIDER_EO_H_
#define _ELM_SLIDER_EO_H_

#ifndef _ELM_SLIDER_EO_CLASS_TYPE
#define _ELM_SLIDER_EO_CLASS_TYPE

typedef Eo Elm_Slider;

#endif

#ifndef _ELM_SLIDER_EO_TYPES
#define _ELM_SLIDER_EO_TYPES


#endif
/** Elementary slider class
 *
 * @ingroup Elm_Slider
 */
#define ELM_SLIDER_CLASS elm_slider_class_get()

EWAPI const Efl_Class *elm_slider_class_get(void) EINA_CONST;

#endif
