/* add mask x color -> dst */

static void
init_add_mask_color_span_funcs_c(void)
{
}

static void
init_add_mask_color_pt_funcs_c(void)
{
}

/*-----*/

/* add_rel mask x color -> dst */

static void
init_add_rel_mask_color_span_funcs_c(void)
{
}

static void
init_add_rel_mask_color_pt_funcs_c(void)
{
}

