#ifndef _EOLIAN_TYPEDEF_EO_STUB_H_STUBS
#define _EOLIAN_TYPEDEF_EO_STUB_H_STUBS

typedef Eo Struct;

/** Docs for struct Named.
 *
 * @since 1.66
 *
 * @ingroup Named
 */
typedef struct _Named Named;

/** Docs for struct Another.
 *
 * @since 1.66
 *
 * @ingroup Another
 */
typedef struct _Another Another;

/** Docs for struct Opaque.
 *
 * @since 1.66
 *
 * @ingroup Opaque
 */
typedef struct _Opaque Opaque;


#endif
