#ifndef ECTOR_SOFTWARE_H_
#define ECTOR_SOFTWARE_H_

#include <Ector.h>

#include "ector_api.h"

#ifdef EFL_BETA_API_SUPPORT

#include "software/ector_software_surface.eo.h"
#include "software/ector_software_buffer.eo.h"
#include "software/ector_software_buffer_base.eo.h"
#include "software/ector_renderer_software.eo.h"
#include "software/ector_renderer_software_shape.eo.h"
#include "software/ector_renderer_software_image.eo.h"
#include "software/ector_renderer_software_gradient_linear.eo.h"
#include "software/ector_renderer_software_gradient_radial.eo.h"

#endif

#endif
