/* sub pixel x color --> dst */

#ifdef BUILD_MMX
static void
init_sub_pixel_color_span_funcs_mmx(void)
{ }
#endif

#ifdef BUILD_MMX
static void
init_sub_pixel_color_pt_funcs_mmx(void)
{ }
#endif

/*-----*/

/* sub_rel pixel x color --> dst */

#ifdef BUILD_MMX
static void
init_sub_rel_pixel_color_span_funcs_mmx(void)
{ }
#endif

#ifdef BUILD_MMX
static void
init_sub_rel_pixel_color_pt_funcs_mmx(void)
{ }
#endif
