#ifndef _EOLIAN_IMPORT_TYPES_EOT_H_
#define _EOLIAN_IMPORT_TYPES_EOT_H_

#ifndef _IMPORT_TYPES_EOT_TYPES
#define _IMPORT_TYPES_EOT_TYPES

typedef int Imported;

typedef struct _Imported_Struct
{
  float foo;
} Imported_Struct;


#endif

#endif
