
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <Eina.h>
#include <Eo.h>

struct Complex_Data {};
typedef struct Complex_Data Complex_Data;

#define EOLIANCXXTEST_API
#define EOLIANCXXTEST_API_WEAK

#include "complex.eo.h"

EOLIAN static void
_complex_inptrcont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inclasscont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_incontcont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_incontcontown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_incontowncontown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_incontowncont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_instringcont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_instringowncont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_instringcontown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inarray(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Array *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inarrayown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Array *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inhash(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Hash *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inhashown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Hash *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_initerator(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Iterator *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_initeratorown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Iterator *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inaccessor(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Accessor *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_inaccessorown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Accessor *l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outclasscont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outcontcont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outcontcontown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outcontowncontown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outcontowncont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outstringcont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outstringowncont(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outstringcontown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outarray(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Array **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outarrayown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Array **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outhash(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Hash **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outhashown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Hash **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outiterator(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Iterator **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outiteratorown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Iterator **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outaccessor(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Accessor **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_outaccessorown(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_Accessor **l EINA_UNUSED)
{
}

EOLIAN static void
_complex_foo(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Eina_List *l EINA_UNUSED)
{
}

EOLIAN static Eina_Array *
_complex_bar(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED)
{
   return NULL;
}

EOLIAN static Complex *
_complex_wrapper_r(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED)
{
   return NULL;
}

EOLIAN static void
_complex_wrapper_in(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Complex *a1 EINA_UNUSED)
{
}

EOLIAN static void
_complex_wrapper_inout(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Complex **a1 EINA_UNUSED)
{
}

EOLIAN static void
_complex_wrapper_out(Eo *obj EINA_UNUSED, Complex_Data *pd EINA_UNUSED, Complex **a1 EINA_UNUSED)
{
}

#include "complex.eo.c"
