#ifndef _ELM_SYS_NOTIFY_DBUS_EO_H_
#define _ELM_SYS_NOTIFY_DBUS_EO_H_

#ifndef _ELM_SYS_NOTIFY_DBUS_EO_CLASS_TYPE
#define _ELM_SYS_NOTIFY_DBUS_EO_CLASS_TYPE

typedef Eo Elm_Sys_Notify_Dbus;

#endif

#ifndef _ELM_SYS_NOTIFY_DBUS_EO_TYPES
#define _ELM_SYS_NOTIFY_DBUS_EO_TYPES


#endif
/** Elementary system notification DBus class
 *
 * @ingroup Elm_Sys_Notify_Dbus
 */
#define ELM_SYS_NOTIFY_DBUS_CLASS elm_sys_notify_dbus_class_get()

EWAPI const Efl_Class *elm_sys_notify_dbus_class_get(void) EINA_CONST;

#endif
