/* this api is currently disabled due to being incomplete. you cannot
 * use it as it is not compiled in */
EAPI Evas_Object                 *elm_factory_add(Evas_Object *parent);
EAPI void                         elm_factory_maxmin_mode_set(Evas_Object *obj, Eina_Bool enabled);
EAPI Eina_Bool                    elm_factory_maxmin_mode_get(const Evas_Object *obj);
EAPI void                         elm_factory_maxmin_reset_set(Evas_Object *obj);
