#ifndef ECORE_EVAS_DRM_H_
# define ECORE_EVAS_DRM_H_

#define EFL_INTERNAL_UNSTABLE
# include "Evas_Internal.h"

typedef struct _Ecore_Evas_Interface_Drm Ecore_Evas_Interface_Drm;

struct _Ecore_Evas_Interface_Drm
{
   Ecore_Evas_Interface base;
};

#endif
